import tokenize as tk
import itertools as it
import unplate.tokenize_util as tku
import unplate.util as util
from unplate.options import Options

"""

Code relating to compiling Python + Unplate source code into native Python source code.

"""


class UnplateSyntaxError(SyntaxError):
  def __init__(self, file_loc, lineno, offset, text, message):
    super().__init__(message)

    # Inherited attributes
    self.file_loc = file_loc  # actually called .filename in SyntaxError
    self.lineno   = lineno
    self.offset   = offset
    self.text     = text

  @property
  def filename(self):
    # SyntaxError expects a .filename, not a .file_loc
    return self.file_loc

  @staticmethod
  def from_token(token, message, *, file_loc=None):
    """ Make an UnplateSyntaxError using the information from a token """

    # Much of the time, the filename won't be given and will
    # be injected higher up in the call stack than the error creation

    return UnplateSyntaxError(
      file_loc = file_loc,
      lineno   = token.start[0],
      offset   = token.start[1],
      text     = token.string,
      message  = message,
    )


def compile_top(tokens, options: Options, *, file_loc):
  """
  Top-level compilation function.
  Transform Python + Unplate source tokens into native Python source tokens.
  """

  try:
    compiled, rest = compile_tokens(tokens, options)
  except UnplateSyntaxError as err:
    err.file_loc = file_loc
    raise err

  if rest:
    raise UnplateSyntaxError.from_token(rest[0], file_loc=file_loc,
      message="Unacceptable leftover tokens.")

  return compiled


def read_string_body(token):
  """
  Given a token denoting a string, return the contents
  of the token without the surrounding quotes.
  So for a token denoting the code `"I am a string"`, return
  the string containing `I am a string`.
  """

  code = token.string

  # strip leading sigils e.g. the 'f' for f-strings
  code = ''.join(it.dropwhile(lambda char: char not in ['"', "'"], code))

  is_multiline_string = len(set(code[:3])) == 1
  if is_multiline_string:
    return code[3:-3]
  else:
    return code[1:-1]


def read_template_body(tokens, indents, options):
  """
  Consume one or more contiguous comments, or a single string.
  Find the contained text.

  In the case of comments, require and consume a leading space
  from the beginning of each comment (e.g.: "# content" -> "content")

  In the case of a string, assert that there are leading and
  trailing lines are empty (whitspace-only), and remove them.
  Additonally, un-indent the contents according to the indent stack.
  """

  if tokens[0].type == tk.STRING:
    content = read_string_body(tokens[0])
    lines = content.split('\n')

    if lines[0].strip() != '':
      raise UnplateSyntaxError.from_token(tokens[0],
        "A template using a Python string literal must begin with a newline.")

    if lines[-1].strip() != '':
      raise UnplateSyntaxError.from_token(tokens[0],
        "A template using a Python string literal must end with a newline.")

    # remove leading blank line
    lines.pop(0)
    # remove trailing blank line
    lines.pop(-1)

    # dedent
    dedented = []
    indent = indents[-1] if indents else ''
    for line in lines:

      # We make a special exception for lines that are only whitespace.
      # We want to allow blank lines to be part of an indended block. However,
      # we don't want to consider all whitespace-only lines as blank lines,
      # in case there is e.g. an embedded multiline string.
      # The compromise is to allow all whitespace-only lines, but 'subtract'
      # the current indent from them, down to a minimum of no indentation.
      if line.strip() == '':
        dedented_line = line[len(indent):]
        dedented.append(dedented_line)

      # normal case:
      # assert then strip indentation
      else:
        if not line.startswith(indent):
          raise UnplateSyntaxError.from_token(tokens[0],
            f"A template using a Python string literal must be indented according to the surrounding block.")
        dedented.append(line[len(indent):])

    return dedented, tokens[1:]

  else:
    comment_block = list(it.takewhile(
      lambda tok: tok.type in [tk.NL, tk.COMMENT],
      tokens
    ))

    comments = [tok for tok in comment_block if tok.type is tk.COMMENT]

    for comment in comments:
      if not comment.string.startswith('# '):
        raise UnplateSyntaxError.from_token(comments[0], "Spaces are mandated after '#' in templates.")

    # [2:] to strip leading space
    lines = [comment.string[2:] for comment in comments]
    rest_tokens = tokens[len(comment_block):]
    return lines, rest_tokens


def repr_with_newlines(string):
  """
  return the string, passed through repr(), but with
  newlines intact.

  Thus the string "new\nline" becomes not "new\\nline" but instead
  a string with a literal newline in it: '''new
line'''
  """

  # single-line string
  if '\n' not in string:
    return repr(string)

  # multiline string
  else:
    reprd = repr(string)

    # either ' or ", depending on how repr() decides to handle escapes
    quote_type = reprd[0]

    newlines_returned = reprd.replace('\\n', '\n')
    multiline_quotes = quote_type * 2 + newlines_returned + quote_type * 2
    return multiline_quotes


def compile_content(string, options):
  """
  Given a string which is the literal content of a template,
  return the Python code for the runtime interpretation of that string.
  For instance, given

    "<h1>{{ title }}</h1>"

  returns (something like)

    "<h1>" + str(title) + "</h1>"

  If the given string contains newlines, these will be preserved in the
  returned code. Thus

    '''first line {{ interpolated }}
    second line'''

  is mapped to (something like)

    "'first line' + str(interpolated) + '''
    second line'''"

  """

  # resultant `exprs` will be a sequence of python expressions which
  # are to be concatenated in the end
  # For example, exprs may be ['"<h1>"', 'str(title)', '"</h1>"']
  exprs = []

  # are we in a string or in an interpolated expression?
  in_expr = False

  # index of the beginning of the current chunk
  # a chunk is either a section of the string which
  # is either literal string or interpolated code
  chunk_start = 0

  def end_chunk():
    chunk = string[chunk_start:i]

    if in_expr:
      code = f"str({chunk})"
    else:
      code = repr_with_newlines(chunk)

    exprs.append(code)

  i = -1
  while True:
    i += 1
    if i >= len(string):
      break

    char = string[i]
    next_char = string[i + 1] if i + 1 < len(string) else None

    if util.starts_with(string, options.interpolation_open, start=i):
      end_chunk()

      # skip over open interpolation
      i += len(options.interpolation_open)
      chunk_start = i

      in_expr = True

    elif in_expr and util.starts_with(string, options.interpolation_close, start=i):
      end_chunk()

      # skip over close interpolation
      i += len(options.interpolation_close)
      chunk_start = i

      in_expr = False

  end_chunk()

  # now we have a bunch of expressions
  # at runtime, they'll need to be joined together
  # we'll do that with ''.join()
  list_expr = '[' + ', '.join(exprs) + ']'
  code = f"''.join({list_expr})"

  return code


def consume_prefix(tokens, literal):
  if not util.prefix_is(tokens, literal):
    expected = tku.untokenize(literal)
    actual = tku.untokenize(tokens[:len(literal)])
    raise UnplateSyntaxError.from_token(tokens[0], f"Expected: {repr(expected)} but got {repr(actual)}")
  return tokens[len(literal):]


def compile_template_literal(tokens, indents, options):
  tokens = consume_prefix(tokens, options.template_literal_open)
  lines, tokens = read_template_body(tokens, indents, options)
  tokens = consume_prefix(tokens, options.template_literal_close)

  content = ''.join(line + '\n' for line in lines)
  compiled = tku.tokenize_expr(compile_content(content, options))

  # Pad compiled code to preserve line numbers
  pad = tku.tokenize_expr('(\n)')
  compiled = pad[:2] + compiled + pad[2:]

  return compiled, tokens


def compile_template_builder(tokens, indents, options):
  """
  Consume and compile a template builder construct ala

    [unplate.begin(template_name)]
    # One line
    # Two line
    # >>> for color in ['red', 'blue']:
      # >>> capitalized = color.upper()
      # {{ color }} line
    # <<<
    [unplate.end]

  Requires the indent stack.
  """

  indents = indents[:]

  compiled = []

  tokens = consume_prefix(tokens, options.template_builder_open_left)
  # get the name of the result template
  template_name = tokens.pop(0).string
  tokens = consume_prefix(tokens, options.template_builder_open_right)

  # consume @ if present
  if tokens[0] == tku.dtok.new(tk.OP, '@'):
    tokens.pop(0)

  init_statement = tku.tokenize_stmt(f"{template_name} = []\n")
  compiled.extend(init_statement)

  # Consume leading newline
  while tokens[0].type == tk.NEWLINE:
    tokens.pop(0)

  body_token = tokens[0]
  lines, tokens = read_template_body(tokens, indents, options)

  # keep track of how many times we've indended in interpolated code
  interpolated_indent_depth = 0

  for line in lines:

    # interpolated python code
    if line.lstrip().startswith('>>>'):

      if not line.lstrip().startswith('>>> '):
        raise UnplateSyntaxError.from_token(body_token, "A space is required after '>>>'")

      python_code = line.lstrip()[len('>>> '):]
      compiled.extend(tku.tokenize_stmt(python_code))
      compiled.append(tku.dtok.new(tk.NEWLINE, '\n'))

      needs_indent = line.strip().endswith(':')
      if needs_indent:
        current_indent = indents[-1] if indents else ''
        bumped = current_indent + '  '
        indents.append(bumped)
        interpolated_indent_depth += 1
        compiled.append(tku.dtok.new(tk.INDENT, bumped))

    elif line.lstrip().startswith('<<<'):

      if line.strip() != '<<<':
        raise UnplateSyntaxError.from_token(body_token, "Nothing is allowed on a line with '<<<'")

      if interpolated_indent_depth == 0:
        raise UnplateSyntaxError.from_token(body_token, "Too many dedents.")

      compiled.append(tku.dtok.new(tk.NEWLINE, '\n'))
      compiled.append(tku.dtok.new(tk.DEDENT, ''))
      indents.pop(-1)
      interpolated_indent_depth -= 1

    else:
      content = tku.tokenize_expr(compile_content(line, options) + " + '\\n'")
      # not entirely sure why the following line needs a trailing \n
      pattern = tku.tokenize_stmt(f'{template_name}.append(VALUE)\n')
      prefix, suffix = tku.split_pattern(pattern, 'VALUE')
      statement = prefix + content + suffix
      compiled.extend(statement)

  # consume the template closing syntax
  tokens = consume_prefix(tokens, options.template_builder_close)

  closing_statement = tku.tokenize_stmt(f"{template_name} = ''.join({template_name})")
  compiled.extend(closing_statement)

  return compiled, tokens


def compile_tokens(tokens, options):
  """
  Given Python tokens that represent Python + Unplate code, compile the Unplate code and return results.
  Results will be a mix of unmodified tokens and raw Python code (as strings).
  """

  compiled = []

  # Keep track of the indentation
  # Each time an indent is reached, push the indentation
  # text (i.e. the actual whitespace) onto this stack
  # Each time a dedent is reached, pop the most recent value
  # off of the stack
  indents = []

  while tokens:
    token = tokens[0]

    if util.prefix_is(tokens, options.template_literal_open):
      compiled_toks, tokens = compile_template_literal(tokens, indents, options)
      compiled.extend(compiled_toks)

    elif util.prefix_is(tokens, options.template_builder_open_left):
      compiled_toks, tokens = compile_template_builder(tokens, indents, options)
      compiled.extend(compiled_toks)

    if token.type == tk.INDENT:
      indents.append(token.string)
      compiled.append(tokens.pop(0))

    elif token.type == tk.DEDENT:
      indents.pop(-1)
      compiled.append(tokens.pop(0))

    else:
      compiled.append(tokens.pop(0))

  return compiled, tokens




def has_sublist(superlist, sublist):
  """
  Does a list contain another list within it?
  Not very efficient.
  If 'eq' is given, this will be used to compare items.
  """
  return any(
    superlist[i : i + len(sublist)] == sublist
    for i in range(len(superlist) - len(sublist) + 1)
  )


def replace_sublist(li, target, replacement):
  """
  Replace a sublist with another sublist.
  Not very effcient.
  If 'eq' is given, this will be used to compare items.
  """
  result = []
  i = 0
  while i < len(li):
    if li[i : i + len(target)] == target:
      i += len(target)
      result.extend(replacement)
    else:
      result.append(li[i])
      i += 1
  return result


def prefix_is(list, prefix):
  return list[:len(prefix)] == prefix


def starts_with(string, substring, *, start: int):
  """
  Functionally equivalent to
    string[start:].startswith(substring)
  but faster
  """
  return string[start : start + len(substring)] == substring



import tokenize as tk
import itertools as it
import operator
import io


class dtok(tk.TokenInfo):
  """
  Represents a "detached" token, which is just
  like a regular token but doesn't care about its
  positional information.

  This entire module uses dtoks instead of
  tokenize.TokenInfo instances.
  """

  def __eq__(self, other):

    if self.type != other.type:
      return False

    # Types for which the content should be compared
    contentful = [
      tk.NAME,
      tk.NUMBER,
      tk.STRING,
      # not 100% sure about these other ones
      #tk.TYPE_COMMENT,
      tk.ERRORTOKEN,
      tk.N_TOKENS,
      tk.NT_OFFSET,
      tk.ENCODING,
    ]

    if self.type in contentful:
      return self.string == other.string
    else:
      return True

  def __str__(self):
    return f"{tk.tok_name[self.type]}({repr(self.string)})"

  def __repr__(self):
    return f"dtok.new({tk.tok_name[self.type]}, {repr(self.string)})"

  @staticmethod
  def new(type, string):
    """
    Construct a dtoken.
    This method exists because __init__ cannot effectively be
    overrid when subclassing a NamedTuple (as far as I know)
    """
    return dtok(type, string, None, None, None)

  @staticmethod
  def from_token(token):
    return dtok(
      token.type,
      token.string,
      token.start,
      token.end,
      token.line,
    )


def tokenize_string(string):
  fake_io = io.StringIO(string)
  tokens = list(tk.generate_tokens(fake_io.readline))
  dtoks = [dtok.from_token(tok) for tok in tokens]
  return dtoks


def tokenize_stmt(string):
  """ Tokenize a string, removing trailing EOF token. """
  tokens = tokenize_string(string)

  assert tokens[-1].type == tk.ENDMARKER
  tokens = tokens[:-1]

  return tokens

def tokenize_expr(string):
  """ Tokenize a string, removing trailing NEWLINE and EOF tokens. """
  tokens = tokenize_string(string)

  assert tokens[-1].type == tk.ENDMARKER
  tokens = tokens[:-1]

  assert tokens[-1].type == tk.NEWLINE
  tokens = tokens[:-1]

  return tokens


def tokenize_one(string):
  """ Tokenize a single token """
  tokens = tokenize_expr(string)
  assert len(tokens) == 1, tokens
  return tokens[0]


def tokenize_file(file_loc):
  """ Tokenize a file """
  with open(file_loc, 'r') as f:
    code = f.read()
  return tokenize_string(code)


def untokenize(tokens):
  return tk.untokenize( (tok.type, tok.string) for tok in tokens )


def split_pattern(pattern_toks, marker):
  """
  Take a pattern and a marker.
  The pattern should be a list of dtoks and the marker should be
  a small bit of Python code (usually a variable name).
  Tokenize both and return the pattern split around the marker.

  For instance:
    split_pattern(tokenize_expr("list[IDX]"), "IDX")
  will return a tuple (prefix, suffix) where
    prefix is the tokens for "list["
    suffix is the tokens for "]"
  """
  marker_tok = tokenize_one(marker)
  marker_idx = pattern_toks.index(marker_tok)
  return pattern_toks[:marker_idx], pattern_toks[marker_idx+1:]



import unplate.compile as unplate_compile
import unplate.tokenize_util as tku
import unplate.util
import unplate.options

# export UnplateSyntaxError
UnplateSyntaxError = unplate_compile.UnplateSyntaxError

def template(*args, **kwargs):
  raise Exception("Something's gone wrong. You should never actually invoke unplate.template().")


true = True


def compile(file_loc, options=options.defaults):

  with open(file_loc, 'r') as f:
    code = f.read()

  return compile_code(code, options, file_loc=file_loc)


def compile_anon(code, options=options.defaults):
  return compile_code(code, options, file_loc='<anonymous>')


def compile_code(code, options=options.defaults, *, file_loc):

  tokens = tku.tokenize_string(code)
  compiled_tokens = unplate_compile.compile_top(tokens, options, file_loc=file_loc)

  # Remove the wrapper
  unplate_wrapper = tku.tokenize_expr('unplate.true')
  true_token = tku.tokenize_one('False')
  compiled_tokens = util.replace_sublist(compiled_tokens, unplate_wrapper, [true_token])

  compiled_code = tku.untokenize(compiled_tokens)
  return compiled_code




import unplate.tokenize_util as tku
import unplate.util

class Options:
  """

  Customizable Unplate parameters.
  Contains the following atributes:

    template_literal_open
      default: the tokens for 'unplate.template('
      The tokens that signifiy the opening of a template literal

    template_literal_close
      default: the tokens for ')'
      The tokens that signify the closing of a template literal

    template_builder_open_left
      default: the tokens for '[unplate.begin('
      The tokens that signify the opening of a template builder, before the variable name

    template_builder_open_right
      default: the tokens for ')'
      The tokens that signify the opening of a template builder, after the variable name

    template_builder_close
      default: the tokens for '[unplate.end]'
      The tokens that signify the closing of a template builder

    interpolation_open
      default: '{{'
      The string that signifies the start of an interpolated Python expression

    interpolation_close
      default: '}}'
      The string that signifies the end of an interpoalted Python expression


  """

  def __init__(self):

    template_literal_pattern = tku.tokenize_expr('unplate.template(\nBODY)')
    self.template_literal_open, self.template_literal_close = tku.split_pattern(template_literal_pattern, 'BODY')

    template_builder_open_pattern = tku.tokenize_expr("[unplate.begin(NAME)]")
    self.template_builder_open_left, self.template_builder_open_right = tku.split_pattern(template_builder_open_pattern, 'NAME')
    self.template_builder_close = tku.tokenize_expr('[unplate.end]')

    self.interpolation_open = '{{'
    self.interpolation_close = '}}'

defaults = Options()



import pytest
import unplate

import unplate.tokenize_util as tku

def test__complain_on_missing_space_after_hash():

  code = """
  templ = unplate.template(
    # space
    #no space
    # space
  )
  """

  with pytest.raises(unplate.UnplateSyntaxError):
    unplate.compile_anon(code)


def test__complain_on_missing_space_after_prompt():

  code = """
  [unplate.begin(template)]
  # >>> 'space'
  # >>>'no space'
  # >>> 'space'
  [unplate.end]
  """

  with pytest.raises(unplate.UnplateSyntaxError):
    unplate.compile_anon(code)


def test__complain_on_non_isolated_dedent():

  code = """
  [unplate.begin(template)]
  # >>> for x in l:
    # >>> pass
  # <<< I should not be here
  [unplate.end]
  """

  with pytest.raises(unplate.UnplateSyntaxError):
    unplate.compile_anon(code)


def test__expr_interpolation():

  code = """#newline
s = "interpolated"
t = unplate.template(
  # {{ s }}
)
assert t == s + '\\n'
  """

  exec(unplate.compile_anon(code))


def test__stmt_interpolation():

  code = """#newline
[unplate.begin(template)]
# first line
# >>> for i in range(3):
  # >>> j = i + 1
  # {{ j }}
# <<<
# last line
[unplate.end]

expected = '''first line
1
2
3
last line
'''

assert template == expected, repr(template)
"""

  exec(unplate.compile_anon(code))


def test__string_templates():

  code = """#newline
[unplate.begin(template)] @ '''
first line
last line
''' [unplate.end]

assert template == 'first line\\nlast line\\n', repr(template)
"""

  exec(unplate.compile_anon(code))


def test__string_templates_indented():

  code = """#newline

if True:
  [unplate.begin(template)] @ '''
  no indent
    one indent
  no indent
  ''' [unplate.end]

  assert template == 'no indent\\n  one indent\\nno indent\\n', repr(template)
"""

  exec(unplate.compile_anon(code))


def test__multiple_indent():

  code = """#newline

if True:
  if True:
    if True:

      [unplate.begin(template)] @ '''
      a
        b
      >>> x = 3
        {{ x }}
      c
      ''' [unplate.end]

      assert template == 'a\\n  b\\n  3\\nc\\n', repr(template)
"""

  exec(unplate.compile_anon(code))


def test__indented_blank_lines():

  code = '''#newline
if cond:
  [unplate.begin(template)] @ """

  """ [unplate.end]
'''

  unplate.compile_anon(code)


def test__unbalanced_indentation():

  code = """#newline
[unplate.begin(template)] @ '''

>>> oops_no_colon
  >>> stmt
<<<

''' [unplate.end]
"""

  with pytest.raises(unplate.UnplateSyntaxError):
    unplate.compile_anon(code)


def test__custom_interpolation():

  options = unplate.options.Options()
  options.interpolation_open = '<PYTHON>'
  options.interpolation_close = '</PYTHON>'

  code = """#newline
x = 3
template = unplate.template(
  # three is <PYTHON> x </PYTHON> is three
)
assert template == 'three is 3 is three\\n', template
"""

  exec(unplate.compile_anon(code, options))


def test__sigils():

  code = """#newline
[unplate.begin(template)] @ r'''
I have a sigil
''' [unplate.end]

assert template == 'I have a sigil\\n', template
"""

  exec(unplate.compile_anon(code))


def test__bug_0001():

  code = '''
if cond:
  [unplate.begin(template)] @ """
  >>> if cond:
        >>> if cond:
          {{ expr }}
        <<<
        {{ expr }}
  <<<
  """ [unplate.end]
'''

  tku.tokenize_string(unplate.compile_anon(code))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name                          = "unplate",
    version                       = "0.2.8",
    author                        = "Quelklef",
    author_email                  = "",
    description                   = "A templating language embedded into Python",
    long_description              = long_description,
    long_description_content_type = "text/markdown",
    url                           = "https://github.com/quelklef/unplate",
    packages                      = setuptools.find_packages(),
    classifiers                   = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires               = '>=3.6',
)
